"""EZyRB metadata."""

__all__ = [
    '__project__', '__title__', '__author__', '__copyright__', '__license__',
    '__version__', '__mail__', '__maintainer__', '__status__'
]

__project__ = 'EZyRB'
__title__ = "ezyrb"
__author__ = "Nicola Demo, Marco Tezzele"
__copyright__ = "Copyright 2016-2021, EZyRB contributors"
__license__ = "MIT"
__version__ = "1.2.1.post2207"
__mail__ = 'demo.nicola@gmail.com, marcotez@gmail.com'
__maintainer__ = __author__
__status__ = "Stable"
